from biobox_analytics.data.adapters._base import Adapter
import _structs as structs
import rdflib
import json
import requests
import math


class GenomeAdapter(Adapter):
    def __init__(self, species, target: structs.Genome ):
        super().__init__()
        self.species = species
        self.assembly = None
        self.__get_taxonid()
        self.__get_ensembl_assembly_info()

    def __get_taxonid(self):
        species = self.species
        # curl request the species name here: curl "https://www.ebi.ac.uk/ena/taxonomy/rest/scientific-name/Leptonycteris%20nivalis"
        url = "https://www.ebi.ac.uk/ena/taxonomy/rest/scientific-name/" + species.replace(" ","%20")
        r = requests.get(url)
        r.json()
        if len(r.json()) == 1:
            self.taxon = int(r.json()[0]['taxId'])

    def __get_ensembl_assembly_info(self):
        species = self.species
        server = "https://rest.ensembl.org"
        ext = "/info/assembly/" + species.replace(" ","_")  + "?"
         
        r = requests.get(server+ext, headers={ "Content-Type" : "application/json"})
         
        if not r.ok:
          r.raise_for_status()
          sys.exit()
         
        decoded = r.json()
        self.assembly = decoded['assembly_name']
        self.karyotypes = decoded['karyotype']
        chromosomes = []
        for chrom in decoded['top_level_region']:
            if (chrom['coord_system'] == "chromosome"):
                chromosomes.append(chrom)
        self.chromosome_regions = chromosomes
                
    def pull_data(self):
        pass

    def iterate_nodes(self):
        # Generate genomic coordinate nodes
        chromosomeRegions = []
        for chrom in self.chromosome_regions:
            maxCoord = math.ceil(chrom['length']/1000)
            for i in range(maxCoord):
                start = (i*1000) + 1
                if (i+1)==(maxCoord):
                    end = chrom['length']
                else:
                    end = (i+1)*1000
                chromosomeRegions.append({
                    "uuid": f"{self.taxon}:{chrom['name']}:{start}-{end}",
                    "displayName": f"{self.species} {chrom}:{start}-{end}"
                    "taxon": self.taxon,
                    "species": self.species,
                    # "assembly": decoded['assembly_name'],
                    "chr": chrom['name'],
                    "start": start,
                    "end": end,
                })
            # write chromosome regions to jsonl file
            
        pass

    def iterate_edges(self):
        # Generate edges between adjacent genomic coordinates
        coordinateEdges = []
        for chrom in self.chromosome_regions:
            maxCoord = math.ceil(chrom['length']/1000)
            # One fewer edge than num nodes
            for i in range(maxCoord-1):
                # current node
                start1 = (i*1000) + 1
                end1 = (i+1)*1000
                start2 = end1 + 1
                if (i+2)==(maxCoord):
                    end2 = chrom['length']
                else:
                    end2 = (i+2)*1000
                # next node
                
                coordinateEdges.append({
                    "uuid": f"9606:{chrom['name']}:{start}-{end}",
                    "taxon": 9606,
                    # "species": "homo sapiens",
                    # "assembly": decoded['assembly_name'],
                    "chrom": chrom['name'],
                    "start": start,
                    "end": end
                })
        pass

    def process_item(self, item):
        """Processes a single item (node or edge)."""
        # Customize how you want to process each item, e.g., add extra information, filter, etc.
        return item  # In this example, we just return the item as-is.

    def describe_node_properties(self):
        pass

    def describe_edge_properties(self):
        pass



from biobox_analytics.data.adapters._base import Adapter
# import biobox_analytics.data.adapters.chipseq._structs as structs
import pandas as pd
import math
import gzip
import os
import json
import datetime
from nanoid import generate

class ChipSeqAdapter(Adapter):
    def __init__(
        self,
        name: str,
        protein_id: str,
        modification_type: str,
        sample_id: str,
        bed_filepath: str,
        taxon_id: str = "9601",
        objects_file: str = "objs.jsonl.gz",
        edges_file: str = "edges.jsonl.gz",
    ):
        super().__init__()
        self.name = name
        self.protein_id = protein_id
        self.modification_type = modification_type
        self.sample_id = sample_id
        self.bed_filepath = bed_filepath
        self.taxon_id = taxon_id
        self.objects_file = objects_file
        self.edges_file = edges_file
        
        self.chipseq_id = generate()

        self.nodes = []
        self.edges = []

    def pull_data(self):
        self.df = pd.read_csv(self.bed_filepath, sep="\t", header=None)
    
    def iterate_nodes(self):
        return self.df.iterrows()
    
    def iterate_edges(self):
        return super().iterate_edges()
    
    def process_item(self, item):
        chr = item.get(0).removeprefix("chr")
        chrom_start = item.get(1)
        chrom_end = item.get(2)
        name = item.get(3)
        score = item.get(4)
        strand = item.get(5)
        signal_value = item.get(6)
        p_value = item.get(7)
        q_value = item.get(8)
        peak = item.get(9)

        narrow_peak_id = self.bed_filepath.removesuffix(".bed") + "-" + name
        labels = [ "NarrowPeak" ]
        properties = {
            "chr": chr,
            "chrom_start": chrom_start,
            "chrom_end": chrom_end,
            # "name": name,
            "score": score,
            "strand": strand,
            "signal_value": signal_value,
            "p_value": p_value,
            "q_value": q_value,
            "peak": peak
        }

        node_narrow_peak = {
            "id": narrow_peak_id,
            "labels": labels,
            "properties": properties
        }

        edge_has_narrow_peak = {
            "label": "has narrow peak",
            "from": {
                "uuid": self.chipseq_id
            },
            "to": {
                "uuid": narrow_peak_id
            }
        }

        range_start = math.floor(chrom_start / 1000) * 1000
        edge_peak_start_on = {
            "label": "peak start on",
            "from": {
                "uuid": narrow_peak_id
            },
            "to": {
                "uuid": "{}:{}:{}:{}".format(self.taxon_id, chr, range_start + 1, range_start + 1000)
            }
        }

        range_end = math.floor(chrom_end / 1000) * 1000
        edge_peak_end_on = {
            "label": "peak end on",
            "from": {
                "uuid": narrow_peak_id
            },
            "to": {
                "uuid": "{}:{}:{}:{}".format(self.taxon_id, chr, range_end + 1, range_end + 1000)
            }
        }


        return [
            node_narrow_peak,
            edge_has_narrow_peak,
            edge_peak_start_on,
            edge_peak_end_on,
        ]
    
    def describe_node_properties(self):
        return super().describe_node_properties()
    
    def describe_edge_properties(self):
        return super().describe_edge_properties()
    
    def extra_items(self):
        chipseq = {
            "id": self.chipseq_id,
            "labels": [ "ChIPseq" ],
            "properties": {
                "uuid": self.chipseq_id,
                "displayName": self.name
            }
        }

        chipseq_to_protein = {
            "label": "assay target on",
            "from": {
                "uuid": self.chipseq_id
            },
            "to": {
                "uuid": self.protein_id
            },
            "properties": {
                "modification_type": self.modification_type
            }
        }
        
        sample_to_chipseq = {
            "label": "has chipseq",
            "from": {
                "uuid": self.sample_id
            },
            "to": {
                "uuid": self.chipseq_id
            },
            # "properties": {
            # }
        }

        return [
            chipseq,
            chipseq_to_protein,
            sample_to_chipseq
        ]
    
    def build(self):
        self.pull_data()
        iterator = self.iterate_nodes()
        for object in self.extra_items():
            print(object)

        for _, row in iterator:
            objects = self.process_item(row)
            for object in objects:
                print(object)
                # if "from" in object:
                #     self.edges.append(object)
                # else:
                #     self.nodes.append(object)

    def write(self):
        obs_file = os.path.join("", self.objects_file)
        edges_file = os.path.join("", self.edges_file)
        with gzip.open(obs_file, "at") as o, gzip.open(edges_file, "at") as e:
            self.pull_data()
            iterator = self.iterate_nodes()
            for object in self.extra_items():
                # print(object)
                if "from" in object:
                    json.dump(object, e)
                    e.write("\n")
                else:
                    json.dump(object, o)
                    o.write("\n")

            for _, row in iterator:
                objects = self.process_item(row)
                for object in objects:
                    # print(object)
                    if "from" in object:
                        json.dump(object, e)
                        e.write("\n")
                    else:
                        json.dump(object, o)
                        o.write("\n")

    def list_schema(self):
        metadata = {
            "_meta": {
                "version": "0.0.1",
                "date_updated": str(datetime.datetime.now()),
            },
            "name": self.name,
            "key": self.name, 
            "description": "",
            "concepts": {
                "NarrowPeak": {
                    "label": "NarrowPeak",
                    "dbLabel": "NarrowPeak",
                    "definition": "",
                },
                "ChIPseq": {
                    "label": "ChIPseq",
                    "dbLabel": "ChIPseq",
                    "definition": "",
                },
            },
            "relationships": {
                "has narrow peak": {
                    "from": "ChIPseq",
                    "to": "NarrowPeak"
                },
                "peak start on": {
                    "from": "NarrowPeak",
                    "to": "GenomicInterval"
                },
                "peak end on": {
                    "from": "NarrowPeak",
                    "to": "GenomicInterval"
                },
                "assay target on": {
                    "from": "ChIPseq",
                    "to": "Protein"
                },
                "has chipseq": {
                    "from": "Sample",
                    "to": "ChIPseq"
                }
            }
        }
        return metadata



from biobox_analytics.data.adapters._base import Adapter
import biobox_analytics.data.adapters.genome._structs as structs
import json
import requests
import os
import gzip
import math
from gtfparse import read_gtf
import urllib.request
import datetime
import itertools

class GenomeAdapter(Adapter):
    def __init__(self, species='homo sapiens', node_filename='node.jsonl.gz', edge_filename='edge.jsonl.gz' ):
        super().__init__()
        self.species = species
        self.assembly = ""
        self.karyotypes = ""
        self.chromosome_regions = []
        self.nodes = []
        self.edges = []
        self.node_filename = node_filename
        self.edge_filename = edge_filename
        self.current_date_time = str(datetime.datetime.now())
        self._gtfloaded = False
        self.taxon = self.__get_taxonid()
        self.__get_ensembl_assembly_info()
        self.displayName = f"Genome Datapack - {self.species} {self.taxon} ({self.current_date_time})"
        self.description = "Genome Datapack created through Python SDK"
        self.key = f"genome:{self.taxon}:{self.current_date_time}"
        # dateCreated = str(datetime.now().isoformat())
        self.genome = structs.Genome(
            uuid=f"genome_{self.assembly}",
            displayName=f"{self.species} - {self.assembly}",
            species=self.species,
            taxon=self.taxon,
            # dateCreated=dateCreated,
            assembly=self.assembly
        )

    def __get_taxonid(self):
        species = self.species
        # curl request the species name here: curl "https://www.ebi.ac.uk/ena/taxonomy/rest/scientific-name/Leptonycteris%20nivalis"
        url = "https://www.ebi.ac.uk/ena/taxonomy/rest/scientific-name/" + species.replace(" ","%20")
        r = requests.get(url)
        r.json()
        if len(r.json()) == 1:
            return int(r.json()[0]['taxId'])
        else:
            print('Taxon not found on species ' + self.species + ' at link ' + url)
            return None

    def __get_ensembl_assembly_info(self, primary_assembly_only=False):
        species = self.species
        server = "https://rest.ensembl.org"
        ext = "/info/assembly/" + species.replace(" ","_")  + "?"
         
        r = requests.get(server+ext, headers={ "Content-Type" : "application/json"})
                  
        decoded = r.json()
        if ('error' in decoded):
            print("Error when fetching ensembl assembly info. " + decoded['error'])
        else:
            self.assembly = decoded['assembly_name']
            self.karyotypes = decoded['karyotype']
            chromosomes = []
            if (primary_assembly_only):
                for chrom in decoded['top_level_region']:
                    if (chrom['coord_system'] == "chromosome"):
                        chromosomes.append(chrom)
            else:
                chromosomes = decoded['top_level_region']
            self.chromosome_regions = chromosomes
                
    def pull_data(self, gtf_url=None):
        if (gtf_url != None):
            url = gtf_url
        elif (self.taxon == 9606):
            url = "https://ftp.ensembl.org/pub/release-112/gtf/homo_sapiens/Homo_sapiens.GRCh38.112.gtf.gz"
        elif (self.taxon == 10090):
            url = "https://ftp.ensembl.org/pub/release-112/gtf/mus_musculus/Mus_musculus.GRCm39.112.gtf.gz"
        else:
            print("Taxon id was not found and gtf_url was not provided")
            return
        
        urllib.request.urlretrieve(url, 'genome.gtf.gz')

    def _generate_genomic_interval_nodes(self):
        genomicIntervals = []
        for chrom in self.chromosome_regions:
            maxCoord = math.ceil(chrom['length']/1000)
            for i in range(maxCoord):
                start = (i*1000) + 1
                # In order to make the end more predictable, the uuid will use (i+1)*1000
                if (i+1)==(maxCoord):
                    end = chrom['length']
                else:
                    end = (i+1)*1000
                genomicIntervals.append({
                    "_id": f"{self.taxon}:{chrom['name']}:{start}-{(i+1)*1000}",
                    "labels": ["GenomicInterval"],
                    "properties": {
                        "uuid": f"{self.taxon}:{chrom['name']}:{start}-{end}",
                        "displayName": f"{self.species} {chrom['name']}:{start}-{end}",
                        "taxon": self.taxon,
                        "species": self.species,
                        # "assembly": decoded['assembly_name'],
                        "chr": chrom['name'],
                        "start": start,
                        "end": end,
                    }
                })
                # chromRegion = structs.GenomicInterval(
                #     taxon=self.taxon,
                #     species=self.species,
                #     uuid=f"{self.taxon}:{chrom['name']}:{start}-{end}",
                #     displayName=f"{self.species} {chrom}:{start}-{end}",
                #     chr=chrom['name'],
                #     start=start,
                #     end=end
                # )
                # genomicIntervals.append({
                #   "_id": f"{self.taxon}:{chrom['name']}:{start}-{end}",
                #   "labels": ["GenomicInterval"],
                #   "properties": chromRegion
                # })
        return genomicIntervals
    
    def _generate_gene_nodes(self):
        genes = self._gtf.filter(feature='gene')
        genes_fixed = []
        for g in genes.iter_rows(named=True):
            name = g['gene_name']
            if name == "":
                name = g['gene_id']
            genes_fixed.append({
                "_id": g['gene_id'],
                "labels": ["Gene"],
                "properties": {
                    "uuid": g['gene_id'],
                    "displayName": name,
                    "chr": g['seqname'],
                    "start": g['start'],
                    "end": g['end'],
                    "assembly": self.assembly,
                    "taxon": self.taxon,
                    "strand": g['strand']
                }
            })
        return genes_fixed
    
    def _generate_transcript_nodes(self):
        transcripts = self._gtf.filter(feature='transcript')
        transcripts_fixed = []
        for t in transcripts.iter_rows(named=True):
            name = t['transcript_name']
            if name == "":
                name = t['transcript_id']
            transcripts_fixed.append({
                "_id": t['transcript_id'],
                "labels": ["Transcript"],
                "properties": {
                    "uuid": t['transcript_id'],
                    "displayName": name,
                    "chr": t['seqname'],
                    "start": t['start'],
                    "end": t['end'],
                    "assembly": self.assembly,
                    "taxon": self.taxon,
                    "strand": t['strand']
                }
            })
        return transcripts_fixed
    
    def _generate_protein_nodes(self):
        protein = self._gtf.filter(self._gtf["protein_id"] != "")
        protein_subset = protein.select(['transcript_id', 'protein_id']).unique()
        protein_distinct = []
        for p in protein_subset.iter_rows(named=True):
            protein_distinct.append({
                "_id": p['protein_id'],
                "labels": ["Protein"],
                "properties": {
                    "uuid": p['protein_id'],
                    "displayName": p['protein_id'],
                    "assembly": self.assembly,
                    "taxon": self.taxon,
                }
            })
        return protein_distinct
    
    def iterate_nodes(self, write_to_disk=True, gtf_file_path = "genome.gtf.gz"):
        if (write_to_disk):
            print(f"Running function in write mode. Writing to file {self.node_filename}. To return nodes, set write_to_disk=False in function call")
            print("Generating genome node")
            nodes = [
                {
                    "_id": self.genome.uuid,
                    "labels": ["Genome"],
                    "properties": {
                        "_id": self.genome.uuid,
                        "displayName": self.genome.displayName,
                        "assembly": self.genome.assembly,
                        "taxon": self.genome.taxon,
                        "chromosomes": json.dumps(self.chromosome_regions)
                    }
                }
            ]
            print("Writing genome node to file")
            self.append_to_file(nodes, filepath=self.node_filename)
            nodes = []
            print("Generating genomic interval nodes")
            nodes = self._generate_genomic_interval_nodes()
            print(f"Writing {len(nodes)} genomic interval nodes to file")
            self.append_to_file(nodes, filepath=self.node_filename)
            nodes = []
            # Load gtf file into dataframe before processing genes, transcripts, proteins
            self._gtf = read_gtf(gtf_file_path)
            self._gtfloaded = True
            print("Generating gene nodes")
            nodes = self._generate_gene_nodes()
            print(f"Writing {len(nodes)} gene nodes to file")
            self.append_to_file(nodes, filepath=self.node_filename)
            nodes = []
            print("Generating transcript nodes")
            nodes = self._generate_transcript_nodes()
            print(f"Writing {len(nodes)} transcript nodes to file")
            self.append_to_file(nodes, filepath=self.node_filename)
            nodes = []
            print("Generating protein nodes")
            nodes = self._generate_protein_nodes()
            print(f"Writing {len(nodes)} protein nodes to file")
            self.append_to_file(nodes, filepath=self.node_filename)
            nodes = []
            print(f"All nodes written to file: {self.node_filename}")
        else:
            print("Running function in non-write mode. Returning nodes. To write to file, set write_to_disk=True in function call")
            genome = [
                {
                    "_id": self.genome.uuid,
                    "labels": ["Genome"],
                    "properties": {
                        "_id": self.genome.uuid,
                        "displayName": self.genome.displayName,
                        "assembly": self.genome.assembly,
                        "taxon": self.genome.taxon,
                        "chromosomes": self.chromosome_regions
                    }
                }
            ]
            print("Generating genomic interval nodes")
            genomic_intervals = self._generate_genomic_interval_nodes()
            # Load gtf file into dataframe before processing genes, transcripts, proteins
            self._gtf = read_gtf(gtf_file_path)
            self._gtfloaded = True
            print("Generating gene nodes")
            genes = self._generate_gene_nodes()
            print("Generating transcript nodes")
            transcripts = self._generate_transcript_nodes()
            print("Generating protein nodes")
            proteins = self._generate_protein_nodes()
            allNodes = [genome, genomic_intervals, genes, transcripts, proteins]
            self.nodes = list(itertools.chain(*allNodes))
            return self.nodes
    
    def _generate_genome_to_genomic_interval_edges(self):
        edges = []
        for chrom in self.chromosome_regions:
            maxCoord = math.ceil(chrom['length']/1000)
            for i in range(maxCoord):
                start = (i*1000) + 1
                # # In order to make uuid's more predictable for genomic intervals, the end is simply (i+1)*1000
                # if (i+1)==(maxCoord):
                #     end = chrom['length']
                # else:
                end = (i+1)*1000
                edges.append({
                    "from": {
                        "uuid": self.genome.uuid,
                    },
                    "to": {
                        "uuid": f"{self.taxon}:{chrom['name']}:{start}-{end}",
                    },
                    "label": "genome contains interval"
                })
        return edges
    
    def _generate_genomic_interval_edges(self):
        # Generate edges between adjacent genomic coordinates
        coordinateEdges = []
        for chrom in self.chromosome_regions:
            maxCoord = math.ceil(chrom['length']/1000)
            # One fewer edge than num nodes
            for i in range(maxCoord-1):
                # current node
                start1 = (i*1000) + 1
                end1 = (i+1)*1000
                start2 = end1 + 1
                # if (i+2)==(maxCoord):
                #     end2 = chrom['length']
                # else:
                # # In order to make uuid's more predictable for genomic intervals, the end is simply (i+2)*1000
                end2 = (i+2)*1000
                # write the edges to a file
                coordinateEdges.append({
                    "from": {
                        "uuid": f"{self.taxon}:{chrom['name']}:{start1}-{end1}",
                    },
                    "to": {
                        "uuid": f"{self.taxon}:{chrom['name']}:{start2}-{end2}",
                    },
                    "label": "next"
                })
        return coordinateEdges
    
    def _generate_gene_transcript_edges(self):
        transcripts = self._gtf.filter(feature='transcript')
        edges = []
        for t in transcripts.iter_rows(named=True):
            edges.append({
                "from": {
                    "uuid": t["gene_id"],
                },
                "to": {
                    "uuid": t["transcript_id"],
                },
                "label": "transcribed to"
            })
        return edges
    
    def _generate_transcript_protein_edges(self):
        protein = self._gtf.filter(self._gtf["protein_id"] != "")
        protein_subset = protein.select(['transcript_id', 'protein_id']).unique()
        edges = []
        for t in protein_subset.iter_rows(named=True):
            edges.append({
                "from": {
                    "uuid": t["transcript_id"],
                },
                "to": {
                    "uuid": t["protein_id"],
                },
                "label": "has translation"
            })
        return edges

    def iterate_edges(self, write_to_disk=True):
        edges = []
        if self._gtfloaded == False:
            print("Run iterate_nodes() to load in the gtf file, prior to this function being callable. Alternatively, call <INSTANCE>._gtf = <INSTANCE>.read_gtf() and <INSTANCE>._gtfloaded = True")
            return
        if (write_to_disk):
            print(f"Running function in write mode. Writing to file {self.edge_filename}. To return edges, set write_to_disk=False in function call")
            print("Processing genome to genomic interval edges")
            edges = self._generate_genome_to_genomic_interval_edges()
            print(f"Writing {len(edges)} genome to genomic interval edges to file")
            self.append_to_file(edges, filepath=self.edge_filename)
            edges = []
            print("Processing genomic interval next edges")
            edges = self._generate_genomic_interval_edges()
            print(f"Writing {len(edges)} genomic interval next edges to file")
            self.append_to_file(edges, filepath=self.edge_filename)
            edges = []
            print("Processing gene to transcript edges")
            edges = self._generate_gene_transcript_edges()
            print(f"Writing {len(edges)} gene to transcript edges to file")
            self.append_to_file(edges, filepath=self.edge_filename)
            edges = []
            print("Processing transcript to protein edges")
            edges = self._generate_transcript_protein_edges()
            print(f"Writing {len(edges)} transcript to protein edges to file")
            self.append_to_file(edges, filepath=self.edge_filename)
            edges = []
            print(f"All edges written to file: {self.edge_filename}")
        else:
            print("Running function in non-write mode. Returning edges. To write to file, set write_to_disk=True in function call")
            print("Processing genome to genomic interval edges")
            gene_genomic_edges = self._generate_genome_to_genomic_interval_edges()
            print("Processing genomic interval next edges")
            genomic_coordinate_edges = self._generate_genomic_interval_edges()
            print("Processing gene to transcript edges")
            gene_transcript_edges = self._generate_gene_transcript_edges()
            transcript_protein_edges = self._generate_transcript_protein_edges()
            print("Processing transcript to protein edges")
            allEdges = [gene_genomic_edges, genomic_coordinate_edges, gene_transcript_edges, transcript_protein_edges]
            print("Returning edges")
            return list(itertools.chain(*allEdges))

    def process_item(self, item):
        """Processes a single item (node or edge)."""
        # Customize how you want to process each item, e.g., add extra information, filter, etc.
        return item  # In this example, we just return the item as-is.

    def describe_node_properties(self):
        pass

    def describe_edge_properties(self):
        pass
    
    def list_schema(self):
        metadata = {
            "_meta": {
                "version": "0.0.1",
                "date_updated": self.current_date_time,
            },
            "name": self.displayName,
            "key": self.key, 
            "description": self.description,
            "concepts": {
                "Gene": {
                    "label": "Gene",
                    "dbLabel": "Gene",
                    "definition": "Gene encompassing all biotypes",
                },
                "Transcript": {
                    "label": "Transcript",
                    "dbLabel": "Transcript",
                    "definition": "Transcripts derived from gene",
                },
                "Protein": {
                    "label": "Protein",
                    "dbLabel": "Protein",
                    "definition": "Protein derived from gene",
                },
                "Genome": {
                    "label": "Genome",
                    "dbLabel": "Genome",
                    "definition": "Genome encompassing this data pack",
                },
                "GenomicInterval": {
                    "label": "Genomic Interval",
                    "dbLabel": "GenomicInterval",
                    "definition": "Genomic Interval splitting the genome's chromosomal regions into sections of 1kbp",
                },
            },
            "relationships": {
                "genome contains interval": {
                    "from": "Genome",
                    "to": "GenomicInterval"
                },
                "next": {
                    "from": "GenomicInterval",
                    "to": "GenomicInterval"
                },
                "transcribed to": {
                    "from": "Gene",
                    "to": "Transcript"
                },
                "has translation": {
                    "from": "Transcript",
                    "to": "Protein"
                },
            }
        }
        return metadata
    
    def append_to_file(self, objs, directory="", filepath="obj.jsonl.gz"):
        with gzip.open(os.path.join(directory, filepath), "at") as f:
            for x in objs:
                json.dump(x, f)
                f.write("\n")

    # def write_serialized_data_to_file(self, directory=""):
    #     node_filepath = os.path.join(directory, self.node_filename)
    #     edge_filepath = os.path.join(directory, self.edge_filename)

    #     with gzip.open(node_filepath, "at") as f:
    #         for x in self.nodes:
    #             json.dump(x, f)
    #             f.write("\n")

    #     with gzip.open(edge_filepath, "at") as f:
    #         for x in self.edges:
    #             json.dump(x, f)
    #             f.write("\n")



from io import StringIO
import requests
import pandas as pd


def fetch_rpt(url: str, headers: list = None) -> pd.DataFrame:
    """
    Fetch a tab-separated file from a given URL, filter out lines starting with '#', and return the content as a pandas DataFrame.

    Parameters:
    url (str): The URL of the file to be fetched.
    headers (list, optional): A list of strings to use as column headers. If not provided, the DataFrame will have default integer column headers.

    Returns:
    pd.DataFrame: A pandas DataFrame containing the filtered content of the file.

    Raises:
    requests.exceptions.RequestException: If the request to the URL fails.
    ValueError: If the content of the response is empty or cannot be parsed into a DataFrame.
    """

    try:
        # Send a GET request
        response = requests.get(url)

        # If the GET request is successful, the status code will be 200
        response.raise_for_status()

        # Get the content of the response
        content = response.text

        if not content:
            raise ValueError("The content of the response is empty.")

        # Filter out the lines that start with '#'
        filtered_content = '\n'.join([line for line in content.split('\n') if not line.startswith('#')])

        # Convert the filtered content into a file-like object
        file_like_object = StringIO(filtered_content)

        # Read the file-like object into a pandas dataframe
        if headers:
            df = pd.read_csv(file_like_object, sep='\t', header=None, names=headers)
        else:
            df = pd.read_csv(file_like_object, sep='\t', header=None)

        return df

    except requests.exceptions.RequestException as e:
        print(f"Request to {url} failed: {e}")
        raise
    except ValueError as e:
        print(f"Error parsing content from {url}: {e}")
        raise



"""
LaminDB x BioBox Integrations
"""

import bionty as bt
from biobox_analytics.core import Node


def _uuid_formatter(db_label, lamin_uid):
    return f'lamindb:{db_label}:{lamin_uid}'


class BxLamin:

    def __init__(self, lamin_obj):
        self._lamin_obj = lamin_obj

        self._db_label = self._get_db_label()
        if self._db_label is None:
            raise Exception("Unrecognized Object Type: This may not be implemented as an integration yet.")

        self._uuid = _uuid_formatter(self._db_label, lamin_obj.uid)

        self._node = self._make_node()

    def save(self):
        self._node.save()
        self._attach_parents()

    def _attach_parents(self):
        parent_df = self._lamin_obj.parents.df()
        parent_records = parent_df.to_dict(orient='records')
        for parent in parent_records:
            if parent.get('ontology_id') is not None:
                self._node.attach('is a', parent.get('ontology_id'))
            else:
                parent_id = _uuid_formatter(self._db_label, parent.get('uid'))
                self._node.attach('is a', parent_id)

    def _is_celltype(self):
        return isinstance(self._lamin_obj, bt.CellType)

    def _get_db_label(self):
        if isinstance(self._lamin_obj, bt.CellType):
            return 'CellType'

        if isinstance(self._lamin_obj, bt.Tissue):
            return 'Tissue'

        if isinstance(self._lamin_obj, bt.Disease):
            return 'Disease'

        if isinstance(self._lamin_obj, bt.ExperimentalFactor):
            return 'EFO'

    def _has_ontology_id(self):
        return self._lamin_obj.ontology_id is not None

    def _make_node(self):
        node = None
        if self._has_ontology_id():
            node = Node(self._lamin_obj.ontology_id, self._db_label)
            node.properties = {'lamindb_xref': self._uuid}
        else:
            node = Node(self._uuid, self._db_label)
            node.properties = {'displayName': self._lamin_obj.name}

        return node


"""
BioBox Concept
"""

import requests
import biobox_analytics._setup as _setup
from urllib.parse import quote
class Concept:

    @classmethod
    def get(cls, db_label):
        res = requests.get(
            f"{_setup.BIOBOX_REST_API}/bioref/concept/{db_label}",
            headers={
                'x-biobox-orgid': _setup.BIOBOX_ORGID,
                'Authorization': f'Bearer {_setup.BIOBOX_TOKEN}'
            }
        )

        res.raise_for_status()

        data = res.json()

        return cls(data)

    def __init__(self, data):
        self._data = data

    @property
    def uri(self):
        return self._data.get('uri')




import json
from abc import ABC, abstractmethod


class Adapter(ABC):
    """
    Abstract base class for an Adapter pattern that processes nodes and edges.

    This class defines a standard interface for pulling data, iterating over nodes
    and edges, processing items, and describing node and edge properties. Subclasses
    should implement the abstract methods to provide specific functionality.
    """

    def __init__(self) -> None:
        """Initialize the Adapter."""
        pass

    def get_nodes(self):
        """
        Iterate over nodes and process each item.

        This method uses `iterate_nodes` to retrieve nodes and processes each node
        using the `process_item` method, then prints the processed result.
        """
        for x in self.iterate_nodes():
            print(self.process_item(x))

    def get_edges(self):
        """
        Iterate over edges and process each item.

        This method uses `iterate_edges` to retrieve edges and processes each edge
        using the `process_item` method, then prints the processed result.
        """
        for x in self.iterate_edges():
            print(self.process_item(x))

    def init(self):
        """Initialize the Adapter. This method can be overridden by subclasses if needed."""
        pass

    def describe(self):
        """
        Describe the properties of nodes and edges.

        This method retrieves descriptions of node and edge properties using
        `describe_node_properties` and `describe_edge_properties` methods and prints
        them as a JSON string.
        """
        node_props = self.describe_node_properties()
        edge_props = self.describe_edge_properties()
        print(json.dumps({"node": node_props, "edge": edge_props}))

    @abstractmethod
    def pull_data(self):
        """Pull data from a data source. Must be implemented by subclasses."""
        pass

    @abstractmethod
    def iterate_nodes(self):
        """Iterate over nodes. Must be implemented by subclasses."""
        pass

    @abstractmethod
    def iterate_edges(self):
        """Iterate over edges. Must be implemented by subclasses."""
        pass

    @abstractmethod
    def process_item(self, item):
        """
        Process a single item (node or edge).

        Parameters:
            item: The item to be processed.

        Returns:
            The processed item.
        """
        pass

    @abstractmethod
    def describe_node_properties(self):
        """
        Describe properties of nodes.

        Returns:
            A description of the node properties.
        """
        pass

    @abstractmethod
    def describe_edge_properties(self):
        """
        Describe properties of edges.

        Returns:
            A description of the edge properties.
        """
        pass



from biobox_analytics.data.adapters._base import Adapter
# import biobox_analytics.data.adapters.scrna._structs as structs
import json
import os
import gzip
import math
import datetime
import itertools
import scanpy
import pandas as pd

class ScRNAAdapter(Adapter):
    def __init__(
        self,
        h5adFile,
        node_filename='node.jsonl.gz',
        edge_filename='edge.jsonl.gz',
    ):
        super().__init__()
        self.input_file = h5adFile
        self.pull_data()
        self.node_filename = node_filename
        self.edge_filename = edge_filename
        self.current_date_time = str(datetime.datetime.now())
        self.displayName = f"SingleCellRNASeq Datapack - {self.current_date_time}"
        self.description = "SingleCellRNASeq Datapack created through Python SDK"
        self.key = f"scrna:{self.current_date_time}"
        self._cellxgeneedges = []
    
    def pull_data(self):
        self.rna = scanpy.read_h5ad(self.input_file)

    def set_metadata(self, displayName, description, key):
        self.displayName = displayName
        self.description = description
        self.key = key
    
    def create_sample_nodes(self, sample_id_col, sample_cols_to_subset):
        samples = []
        for index, row in self.rna.obs[sample_cols_to_subset].drop_duplicates(subset=sample_id_col).iterrows():
            props = row.to_dict()
            props['displayName'] = row[sample_id_col]
            samples.append({
                "_id": row[sample_id_col],
                "labels": ["Sample"],
                "properties": props
            })
        return samples
    
    def create_sc_experiment_library_nodes(self, sc_library_experiment_id, library_experiment_cols_to_subset):
        experiments = []
        for index, row in self.rna.obs[library_experiment_cols_to_subset].drop_duplicates(subset=sc_library_experiment_id).iterrows():
            props = row.to_dict()
            props['displayName'] = row[sc_library_experiment_id]
            experiments.append({
                "_id": row[sc_library_experiment_id],
                "labels": ["Experiment","SingleCellExperiment","SingleCellRNAseqExperiment"],
                "properties": props
            })
        return experiments
    
    def create_cell_nodes(self, sc_library_experiment_id):
        barcodes = []
        for index, row in self.rna.obs.iterrows():
            barcodes.append({
                "_id": f"{row[sc_library_experiment_id]}:{index}",
                "labels": ["CellBarcode"],
                "properties": {
                    "displayName": f"{row[sc_library_experiment_id]}:{index}"
                }
            })
        return barcodes

    def iterate_nodes(self, write_to_disk=True, sc_library_experiment_id="library_uuid", sample_id_col="sample_uuid", sc_experiment_cols_to_subset=["library_uuid"], sample_metadata_cols_to_subset=["sample_uuid"]):
        if (write_to_disk):
            print(f"Running function in write mode. Writing to file {self.node_filename}. To return nodes, set write_to_disk=False in function call")
            nodes = []
            print(f"Create cell nodes")
            nodes = self.create_cell_nodes(sc_library_experiment_id)
            print(f"Writing {len(nodes)} cell nodes to file")
            self.append_to_file(nodes, filepath=self.node_filename)
            nodes = []
            print(f"Create experiment nodes")
            nodes = self.create_sc_experiment_library_nodes(sc_library_experiment_id, sc_experiment_cols_to_subset)
            print(f"Writing {len(nodes)} experiment nodes to file")
            self.append_to_file(nodes, filepath=self.node_filename)
            if (sample_id_col):
                nodes = []
                print(f"Create sample nodes")
                nodes = self.create_sample_nodes(sample_id_col, sample_metadata_cols_to_subset)
                print(f"Writing {len(nodes)} sample nodes to file")
                self.append_to_file(nodes, filepath=self.node_filename)
                nodes = []
            print(f"All nodes written to file: {self.node_filename}")
        else:
            print("Running function in non-write mode. Returning nodes. To write to file, set write_to_disk=True in function call")
            barcodes = self.create_cell_nodes(sc_library_experiment_id)
            singleCellExperiments = self.create_sc_experiment_library_nodes(sc_library_experiment_id, sc_experiment_cols_to_subset)
            samples = []
            if (sample_id_col):
                samples = self.create_sample_nodes(sample_id_col, sample_metadata_cols_to_subset)
            return list(itertools.chain(*[barcodes, singleCellExperiments, samples]))
    
    def create_sample_to_experiment_connection(self, sc_library_experiment_id, sample_id_col):
        edges = []
        for index, row in self.rna.obs.drop_duplicates(subset=sc_library_experiment_id).iterrows():
            edges.append({
                "from": {
                    "uuid": row[sample_id_col]
                },
                "to": {
                    "uuid": row[sc_library_experiment_id]
                },
                "label": "has experiment"
            })
        return edges
    
    def create_experiment_to_barcode_connection(self, sc_library_experiment_id):
        edges = []
        for index, row in self.rna.obs.iterrows():
            edges.append({
                "from": {
                    "uuid": row[sc_library_experiment_id]
                },
                "to": {
                    "uuid": f"{row[sc_library_experiment_id]}:{index}"
                },
                "label": "contains cell"
            })
        return edges
    
    def create_barcode_to_celltype_connection(self, sc_library_experiment_id, celltype_id_col):
        edges = []
        for index, row in self.rna.obs.iterrows():
            edges.append({
                "from": {
                    "uuid": f"{row[sc_library_experiment_id]}:{index}"
                },
                "to": {
                    "uuid": row[celltype_id_col]
                },
                "label": "has cell type"
            })
        return edges

    def _process_barcodes_genes(self, row, barcode):
        return {
            "from": {
                "uuid": barcode
            },
            "to": {
                "uuid": row.name
            },
            "label": "expresses",
            "properties": {
                "normValue": row[barcode].astype(float)
            }
        }
    
    def _process_barcode(self, column):
        barcode = column.name
        cell = pd.DataFrame(column)
        filtered_cell = cell[cell[column.name] > 0]
        # cell2 = cell.set_axis(['expression'], axis=1)
        a = filtered_cell.apply(lambda row: self._process_barcodes_genes(row, barcode), axis=1)
        self._cellxgeneedges.extend(a.tolist())
    
    def create_barcode_to_gene_connection(self, sc_library_experiment_id):
        numcells = self.rna.X.shape[0]
        print(f"Number of cells to process: {numcells}")
        print(f"Starting Cell x Gene edge processing now: {datetime.datetime.now()}")
        for i in range(math.ceil(self.rna.X.shape[0]/1000)):
            low = i*1000
            high = (i+1)*1000
            if (high > numcells):
                high = numcells
            print(f"Processing batch index {low}:{high} at time {datetime.datetime.now()}")
            library = self.rna.obs[sc_library_experiment_id][low:high]
            cell = self.rna.obs.index[low:high]
            cell_uniq = [f"{libi}:{celli}" for libi, celli in zip(library.tolist(), cell.tolist())]
            rowPandas = pd.DataFrame(self.rna.X[low:high].toarray().T, index=self.rna.var.index, columns=cell_uniq)
            rowPandas.apply(self._process_barcode, axis=0)
            self.append_to_file(self._cellxgeneedges, filepath=self.edge_filename)
            self._cellxgeneedges = []
        print(f"Cell x Gene edges written to file: {self.edge_filename} at time {datetime.datetime.now()}")
        return []

    def iterate_edges(self, write_to_disk=True, sc_library_experiment_id="library_uuid", sample_id_col=None, celltype_id_col=None, tissuetype_id_col=None):
        if (write_to_disk):
            print(f"Running function in write mode. Writing to file {self.edge_filename}. To return edges, set write_to_disk=False in function call")
            edges = []
            print(f"Calculating experiment-cell edges")
            edges = self.create_experiment_to_barcode_connection(sc_library_experiment_id)
            print(f"Writing {len(edges)} experiment-cell edges to file")
            self.append_to_file(edges, filepath=self.edge_filename)
            if sample_id_col != None:
                print(f"Calculating sample-experiment edges")
                edges = self.create_sample_to_experiment_connection(sc_library_experiment_id, sample_id_col)
                print(f"Writing {len(edges)} sample-experiment edges to file")
                self.append_to_file(edges, filepath=self.edge_filename)
                edges = []
            if celltype_id_col != None:
                print(f"Calculating cell-celltype edges")
                edges = self.create_barcode_to_celltype_connection(sc_library_experiment_id, celltype_id_col)
                print(f"Writing {len(edges)} cell-celltype edges to file")
                self.append_to_file(edges, filepath=self.edge_filename)
                edges = []
            print(f"Calculating cell-gene edges")
            self.create_barcode_to_gene_connection(sc_library_experiment_id)
            print(f"All edges written to file: {self.edge_filename}")
        else:
            print("Running function in non-write mode. Returning edges. To write to file, set write_to_disk=True in function call")
            exp_to_barcode_edges = self.create_experiment_to_barcode_connection(sc_library_experiment_id)
            print("Skipping cell-gene edges as the matrix is too large")
            # barcode_to_gene_edges = self.create_barcode_to_gene_connection(sc_library_experiment_id)
            sample_exp_edges = []
            if sample_id_col != None:
                sample_exp_edges = self.create_sample_to_experiment_connection(sc_library_experiment_id, sample_id_col)
            barcode_celltype_edges = []
            if self.enable_celltype_connections:
                barcode_celltype_edges = self.create_barcode_to_celltype_connection(sc_library_experiment_id, celltype_id_col)
            return list(itertools.chain(*[exp_to_barcode_edges, barcode_celltype_edges, sample_exp_edges]))

    def process_item(self, item):
        """Processes a single item (node or edge)."""
        # Customize how you want to process each item, e.g., add extra information, filter, etc.
        return item  # In this example, we just return the item as-is.

    def describe_node_properties(self):
        pass

    def describe_edge_properties(self):
        pass
    
    def list_schema(self):
        metadata = {
            "_meta": {
                "version": "0.0.1",
                "date_updated": self.current_date_time,
            },
            "name": self.displayName,
            "key": self.key, 
            "description": self.description,
            "dependencies": ["Ensembl"],
            "concepts": {
                "Experiment": {
                    "label": "Experiment",
                    "dbLabel": "Experiment",
                    "definition": "Experiment of the sample tissue",
                },
                "SingleCellExperiment": {
                    "label": "SingleCellExperiment",
                    "dbLabel": "SingleCellExperiment",
                    "definition": "Single Cell Experiment of the sample tissue",
                    "sco": "Experiment",
                },
                "SingleCellRNAseqExperiment": {
                    "label": "SingleCellRNAseqExperiment",
                    "dbLabel": "SingleCellRNAseqExperiment",
                    "definition": "Single Cell RNAseq Experiment of the sample tissue",
                    "sco": "SingleCellExperiment",
                },
                "Sample": {
                    "label": "Sample",
                    "dbLabel": "Sample",
                    "definition": "Sample organism from which tissue was taken to be analyzed",
                },
                "CellBarcode": {
                    "label": "Cell Barcode",
                    "dbLabel": "CellBarcode",
                    "definition": "Individual cell from scRNA experiment, identified by barcode",
                }
            },
            "relationships": {
                "contains cell": {
                    "from": "SingleCellExperiment",
                    "to": "CellBarcode"
                },
                "expresses": {
                    "from": "CellBarcode",
                    "to": "Gene"
                },
                "has experiment": {
                    "from": "Sample",
                    "to": "Experiment"
                },
                "has cell type": {
                    "from": "CellBarcode",
                    "to": "CellType"
                },
            }
        }
        return metadata
    
    def append_to_file(self, objs, directory="", filepath="obj.jsonl.gz"):
        with gzip.open(os.path.join(directory, filepath), "at") as f:
            for x in objs:
                json.dump(x, f)
                f.write("\n")

    # def write_serialized_data_to_file(self, directory=""):
    #     node_filepath = os.path.join(directory, self.node_filename)
    #     edge_filepath = os.path.join(directory, self.edge_filename)

    #     with gzip.open(node_filepath, "at") as f:
    #         for x in self.nodes:
    #             json.dump(x, f)
    #             f.write("\n")

    #     with gzip.open(edge_filepath, "at") as f:
    #         for x in self.edges:
    #             json.dump(x, f)
    #             f.write("\n")



import os
import json
from tqdm import tqdm
import requests
from concurrent.futures import ThreadPoolExecutor
import io
import gzip

def setup_working_directories(base_path):
    tmp_data = os.path.join(base_path, 'tmp_data')
    processed_data = os.path.join(base_path, 'processed_data')

    os.makedirs(tmp_data, exist_ok=True)
    os.makedirs(processed_data, exist_ok=True)

def ensure_primitive_or_array_of_primitives(value):
    if isinstance(value, (str, int, float, bool)) or value is None:
        return value
    elif isinstance(value, list):
        return [ensure_primitive_or_array_of_primitives(v) for v in value]
    else:
        return json.dumps(value)


def read_jsonl_files(directory):
    """
    Recursively reads all .jsonl files in a directory and its subdirectories.

    Args:
        directory (str): The path to the directory to scan for .jsonl files.

    Returns:
        list: A list of dictionaries, where each dictionary represents the contents of a .jsonl file.
    """
    data = []  # Initialize an empty list to store the data
    jsonl_files = []  # Initialize an empty list to store the .jsonl files
    
    # Find all .jsonl files in the directory and its subdirectories
    for root, dirs, files in os.walk(directory):
        for file in files:
            if file.endswith(".jsonl") or file.endswith(".json"):
                jsonl_files.append(os.path.join(root, file))
    
    # Read the contents of each .jsonl file
    for file_path in tqdm(jsonl_files, desc="Loading .jsonl files"):
        with open(file_path, 'r') as f:
            # Read the file line by line
            for line in f:
                # Parse each line as JSON and add it to the data list
                data.append(json.loads(line))
    
    return data

def download_gzipped_json(uri):
    # Send a GET request to the URI
    response = requests.get(uri)

    # Check if the response was successful
    if response.status_code != 200:
        raise Exception(f"Failed to download file from {uri}")

    # Create a BytesIO object to hold the gzipped data
    gzipped_data = io.BytesIO(response.content)

    # Create a GzipFile object to decompress the data
    with gzip.GzipFile(fileobj=gzipped_data, mode='rb') as f:
        # Read the decompressed data into a string
        json_data = f.read().decode('utf-8')

    # Parse the JSON data into a Python object
    data = json.loads(json_data)

    return data


def oxo_mapping(ids, target_ontologies):
    url = "https://www.ebi.ac.uk/spot/oxo/api/search"
    json_body = {
        'ids': ids,
        'mappingTarget': target_ontologies,
        'distance': 3
    }
    initial_request = requests.post(
        url=url,
        json=json_body
    )
    initial_request.raise_for_status()
    data = initial_request.json()
    total_pages = data.get('page', {}).get('totalPages')
    results = {}
    for i in tqdm(range(total_pages)):
        res = requests.post(
            url=url,
            json=json_body
        )
        data = res.json()
        mappings = data.get('_embedded', {}).get('searchResults', [])

        for m in mappings:
            mappingResponseList = m.get('mappingResponseList', [])
            if len(mappingResponseList) == 0:
                results[m.get('queryId')] = None
            else:
                mapping_prioritized = sorted(mappingResponseList,
                                             key=lambda x: (x['targetPrefix'] != 'EFO', x['distance']))
                results[m.get('queryId')] = [x.get('curie') for x in mapping_prioritized]

        next_url = data.get('_links', {}).get('next', {}).get('href')
        if next_url is None:
            break
        else:
            url = next_url
    return results



def scrape_page(url, page, limit):
    """
    Scrape a single page of records from the API.

    Args:
        url (str): The URL of the API endpoint.
        page (int): The page number to scrape.
        limit (int): The number of records per page.

    Returns:
        list: A list of records from the specified page.
    """
    params = {'page': page, 'limit': limit}
    response = requests.get(url, params=params)
    data = response.json()
    return data['results']

def scrape_all_records(url, limit=20, max_workers=10):
    """
    Scrape all records from the API.

    Args:
        url (str): The URL of the API endpoint.
        limit (int, optional): The number of records per page. Defaults to 20.
        max_workers (int, optional): The maximum number of concurrent requests. Defaults to 10.

    Returns:
        list: A list of all records from the API.
    """
    response = requests.get(url, params={'page': 1, 'limit': limit})
    data = response.json()
    total_pages = -(-data['total'] // limit)  # ceiling division

    with ThreadPoolExecutor(max_workers=max_workers) as executor:
        futures = [executor.submit(scrape_page, url, page, limit) for page in range(1, total_pages + 1)]
        all_records = [record for future in futures for record in future.result()]

    return all_records


"""
BioBox Node
"""
import requests
import biobox_analytics._setup as _setup
from biobox_analytics.core._concept import Concept
from urllib.parse import quote
import json
class Node:

    def __init__(self, uuid, db_label):
        self.uuid = uuid
        self.db_label = db_label
        bx_data = self._load()

        self._exists = False

        if bx_data is not None:
            self._exists = True
            self._data = bx_data
        else:
            self._data = {
                'uuid': uuid,
                'displayName': 'Node'
            }
        self.concept = Concept.get(db_label)

    @property
    def displayName(self):
        return self._data.get('displayName')

    @displayName.setter
    def displayName(self, v):
        self._data['displayName'] = v

    @property
    def properties(self):
        return self._data['properties']

    @properties.setter
    def properties(self, adict):
        self._data['properties'].update(adict)

    def save(self):
        if self._exists:
            self._update()
        else:
            self._create()

        print(json.dumps(self._data, indent=2))

    def _update(self):
        res = requests.put(
            f"{_setup.BIOBOX_REST_API}/bioref/object/{quote(self.uuid)}",
            json={
                'uuid': self.uuid,
                'displayName': self.displayName
            },
            headers={
                'x-biobox-orgid': _setup.BIOBOX_ORGID,
                'Authorization': f'Bearer {_setup.BIOBOX_TOKEN}'
            }
        )
        res.raise_for_status()
        return res.json()

    def delete(self):
        pass

    def _load(self):
        res = requests.get(
            f"{_setup.BIOBOX_REST_API}/bioref/object/{quote(self.uuid)}",
            headers={
                'x-biobox-orgid': _setup.BIOBOX_ORGID,
                'Authorization': f'Bearer {_setup.BIOBOX_TOKEN}'
            }
        )
        if res.status_code == 200:
            return res.json()

    def _create(self):
        res = requests.post(
            f"{_setup.BIOBOX_REST_API}/bioref/object",
            json={
                'conceptUri': self.concept.uri,
                'payload': [
                    {
                        'uuid': self.uuid,
                        'displayName': self.displayName,
                        'properties': [
                            {
                                'dataPropertyUri': 'uuid',
                                'value': self.uuid
                            },
                            {
                                'dataPropertyUri': 'displayName',
                                'value': self.displayName
                            }
                        ]
                    }
                ]
            },
            headers={
                'x-biobox-orgid': _setup.BIOBOX_ORGID,
                'Authorization': f'Bearer {_setup.BIOBOX_TOKEN}'
            }
        )
        res.raise_for_status()
        return res.json()

    def attach(self, rel, target):
        res = requests.post(
            f"{_setup.BIOBOX_REST_API}/bioref/attach/object",
            json={
                'from': {
                    'uuid': self.uuid
                },
                'to': {
                    'uuid': target
                },
                'label': rel,
                'properties': []
            },
            headers={
                'x-biobox-orgid': _setup.BIOBOX_ORGID,
                'Authorization': f'Bearer {_setup.BIOBOX_TOKEN}'
            }
        )
        if res.status_code != 200:
            print(res.text)
        res.raise_for_status()
        return res.json()



"""
BioBox Core
"""

from biobox_analytics.core._concept import Concept
from biobox_analytics.core._node import Node

from biobox_analytics.data.adapters.genome import GenomeAdapter
from biobox_analytics.data.adapters.scrna import ScRNAAdapter
from biobox_analytics.data.adapters.scatac import ScATACAdapter
from biobox_analytics.data.adapters.chipseq import ChipSeqAdapter

__all__ = ['GenomeAdapter', 'ScRNAAdapter', 'ScATACAdapter', 'ChipSeqAdapter']


from enum import Enum, auto
import requests
import os
import json
from biobox_analytics.utils import download_gzipped_json, oxo_mapping, scrape_all_records
import io
import pandas as pd
from tqdm import tqdm


class AllianceDownloads(Enum):
    DISEASE_HUMAN = "https://fms.alliancegenome.org/download/DISEASE-ALLIANCE-JSON_HUMAN.json.gz"
    DISEASE_MOUSE = "https://fms.alliancegenome.org/download/DISEASE-ALLIANCE-JSON_MGI.json.gz"


class AllianceGenomeAdapter:

    def __init__(self):
        self._api = "https://www.alliancegenome.org/api/"
        self._disease_gene_associations = []
        self._prepare_mappers()
        self._download_disease_associations()

        self.metadata = {
            '_meta': {
                'version': '0.1.0',
                'date_updated': '2024-05-19',
                'maintainer': 'BioBox Analytics'
            },
            'key': 'alliance_genome',
            'name': 'Alliance of Genome Resources',
            'description': 'TBD',
            'source': [
                {
                    'uri': 'https://www.alliancegenome.org/',
                    'type': 'doc'
                },
                {
                    'uri': 'https://www.alliancegenome.org/downloads',
                    'type': 'data',
                    'version': '7.1.0'
                }
            ],
            'concepts': {},
            'relationships': {}
        }

    def _prepare_mappers(self):

        human = pd.read_table("https://www.alliancegenome.org/api/geneMap/ensembl?species=NCBITaxon:9606", sep="\t")

        mouse = pd.read_table("https://www.alliancegenome.org/api/geneMap/ensembl?species=NCBITaxon:10090", sep="\t")

        self._gene_map = pd.concat([human, mouse])

    def _download_disease_associations(self):
        human_results = download_gzipped_json(AllianceDownloads.DISEASE_HUMAN.value)
        DISEASE_IDS = list(set([x.get('DOID') for x in human_results['data']]))

        self._doid_mapping = oxo_mapping(DISEASE_IDS, ["EFO", "MONDO"])

        for _id in tqdm(DISEASE_IDS, desc="Collecting DiseaseGene Annotations"):
            try:
                harmonized_disease_id = self._doid_mapping.get(_id)[0]
            except:
                continue

            records = scrape_all_records(
                f"https://www.alliancegenome.org/api/disease/{_id}/genes?filter.species=Homo sapiens")
            associations = []
            for anno in records:
                association = self._format_disease_gene_annotation(anno, harmonized_disease_id)
                if association is not None:
                    associations.append(association)
            self._disease_gene_associations.extend(associations)

    def _hgnc2ensembl(self, hgnc):
        hgnc_id = hgnc.split(':').pop()

        subset = self._gene_map[self._gene_map['NCBI ID'] == hgnc_id]
        return subset['Ensembl ID']

    def _format_disease_gene_annotation(self, anno, disease_id, fmt="edge"):

        relation_str = anno.get('generatedRelationString').replace('_', ' ')
        disease_qualifiers = anno.get('diseaseQualifiers', [])

        if len(disease_qualifiers) > 1:
            print(anno)
            raise Exception("Disease Qualifiers should not be more than one in length")

        if len(disease_qualifiers) == 1:
            disease_qualifier_str = disease_qualifiers[0].replace('_', ' ')
            relation_str = f'{relation_str} {disease_qualifier_str}'

        uuid = anno.get("uniqueId")

        references = anno.get('pubmedPubModIDs')
        evidence = map(lambda x: x.get('abbreviation'), anno.get('evidenceCodes', []))

        try:
            gene_hgnc = anno['subject']['modEntityId']
            ensembl_id = self._hgnc2ensembl(gene_hgnc)
        except:
            return

        output = {
            'from': {
                'uuid': ensembl_id
            },
            'to': {
                'uuid': disease_id
            },
            'label': relation_str,
            'properties': {
                'uuid': uuid,
                'references': references,
                'evidence': evidence,
                'citation_authority': 'Alliance of Genome Resources'
            }
        }
        return output



from biobox_analytics.data.adapters._base import Adapter
import biobox_analytics.data.adapters.genome._structs as structs
import json
import requests
import os
import gzip
import math
from gtfparse import read_gtf
import urllib.request
from datetime import datetime
import itertools

class ScATAC(Adapter):
    def __init__(self, experimentID, experimentName, cellByPeakPath, celltypes=None, node_filename='node.jsonl.gz', edge_filename='edge.jsonl.gz' ):
        super().__init__()
        self.experiment = structs.SingleCellATACseqExperiment(
            uuid=experimentID,
            displayName=experimentName
        )
        self.node_filename = node_filename
        self.edge_filename = edge_filename

    def _generate_genomic_interval_nodes(self):
        genomicIntervals = []
        for chrom in self.chromosome_regions:
            maxCoord = math.ceil(chrom['length']/1000)
            for i in range(maxCoord):
                start = (i*1000) + 1
                if (i+1)==(maxCoord):
                    end = chrom['length']
                else:
                    end = (i+1)*1000
                genomicIntervals.append({
                    "_id": f"{self.taxon}:{chrom['name']}:{start}-{end}",
                    "labels": ["GenomicInterval"],
                    "properties": {
                        "uuid": f"{self.taxon}:{chrom['name']}:{start}-{end}",
                        "displayName": f"{self.species} {chrom['name']}:{start}-{end}",
                        "taxon": self.taxon,
                        "species": self.species,
                        # "assembly": decoded['assembly_name'],
                        "chr": chrom['name'],
                        "start": start,
                        "end": end,
                    }
                })
                # chromRegion = structs.GenomicInterval(
                #     taxon=self.taxon,
                #     species=self.species,
                #     uuid=f"{self.taxon}:{chrom['name']}:{start}-{end}",
                #     displayName=f"{self.species} {chrom}:{start}-{end}",
                #     chr=chrom['name'],
                #     start=start,
                #     end=end
                # )
                # genomicIntervals.append({
                #   "_id": f"{self.taxon}:{chrom['name']}:{start}-{end}",
                #   "labels": ["GenomicInterval"],
                #   "properties": chromRegion
                # })
        return genomicIntervals
    
    def _generate_gene_nodes(self):
        genes = self._gtf.filter(feature='gene')
        genes_fixed = []
        for g in genes.iter_rows(named=True):
            name = g['gene_name']
            if name == "":
                name = g['gene_id']
            genes_fixed.append({
                "_id": g['gene_id'],
                "labels": ["Gene"],
                "properties": {
                    "uuid": g['gene_id'],
                    "displayName": name,
                    "chr": g['seqname'],
                    "start": g['start'],
                    "end": g['end'],
                    "assembly": self.assembly,
                    "taxon": self.taxon,
                    "strand": g['strand']
                }
            })
        return genes_fixed
    
    def _generate_transcript_nodes(self):
        transcripts = self._gtf.filter(feature='transcript')
        transcripts_fixed = []
        for t in transcripts.iter_rows(named=True):
            name = t['transcript_name']
            if name == "":
                name = t['transcript_id']
            transcripts_fixed.append({
                "_id": t['transcript_id'],
                "labels": ["Transcript"],
                "properties": {
                    "uuid": t['transcript_id'],
                    "displayName": name,
                    "chr": t['seqname'],
                    "start": t['start'],
                    "end": t['end'],
                    "assembly": self.assembly,
                    "taxon": self.taxon,
                    "strand": t['strand']
                }
            })
        return transcripts_fixed
    
    def _generate_protein_nodes(self):
        protein = self._gtf.filter(self._gtf["protein_id"] != "")
        protein_subset = protein.select(['transcript_id', 'protein_id']).unique()
        protein_distinct = []
        for p in protein_subset.iter_rows(named=True):
            protein_distinct.append({
                "_id": p['protein_id'],
                "labels": ["Protein"],
                "properties": {
                    "uuid": p['protein_id'],
                    "displayName": p['protein_id'],
                    "assembly": self.assembly,
                    "taxon": self.taxon,
                }
            })
        return protein_distinct
    
    def iterate_nodes(self, gtf_file_path = "genome.gtf.gz"):
        genome = [
            {
                "_id": self.genome.uuid,
                "labels": ["Genome"],
                "properties": {
                    "_id": self.genome.uuid,
                    "displayName": self.genome.displayName,
                    "assembly": self.genome.assembly,
                    "taxon": self.genome.taxon,
                    "chromosomes": self.chromosome_regions
                }
            }
        ]
        genomic_intervals = self._generate_genomic_interval_nodes()
        # Load gtf file into dataframe before processing genes, transcripts, proteins
        self._gtf = read_gtf(gtf_file_path)
        self._gtfloaded = True
        genes = self._generate_gene_nodes()
        transcripts = self._generate_transcript_nodes()
        proteins = self._generate_protein_nodes()
        allNodes = [genome, genomic_intervals, genes, transcripts, proteins]
        self.nodes = list(itertools.chain(*allNodes))
        return self.nodes
    
    def _generate_genome_to_genomic_interval_edges(self):
        edges = []
        for chrom in self.chromosome_regions:
            maxCoord = math.ceil(chrom['length']/1000)
            for i in range(maxCoord):
                start = (i*1000) + 1
                if (i+1)==(maxCoord):
                    end = chrom['length']
                else:
                    end = (i+1)*1000
                edges.append({
                    "from": {
                        "uuid": self.genome.uuid,
                    },
                    "to": {
                        "uuid": f"{self.taxon}:{chrom['name']}:{start}-{end}",
                    },
                    "label": "genome contains interval"
                })
        return edges
    
    def _generate_genomic_interval_edges(self):
        # Generate edges between adjacent genomic coordinates
        coordinateEdges = []
        for chrom in self.chromosome_regions:
            maxCoord = math.ceil(chrom['length']/1000)
            # One fewer edge than num nodes
            for i in range(maxCoord-1):
                # current node
                start1 = (i*1000) + 1
                end1 = (i+1)*1000
                start2 = end1 + 1
                if (i+2)==(maxCoord):
                    end2 = chrom['length']
                else:
                    end2 = (i+2)*1000
                # write the edges to a file
                coordinateEdges.append({
                    "from": {
                        "uuid": f"{self.taxon}:{chrom['name']}:{start1}-{end1}",
                    },
                    "to": {
                        "uuid": f"{self.taxon}:{chrom['name']}:{start2}-{end2}",
                    },
                    "label": "next"
                })
        return coordinateEdges
    
    def _generate_gene_transcript_edges(self):
        transcripts = self._gtf.filter(feature='transcript')
        edges = []
        for t in transcripts.iter_rows(named=True):
            edges.append({
                "from": {
                    "uuid": t["gene_id"],
                },
                "to": {
                    "uuid": t["transcript_id"],
                },
                "label": "transcribed to"
            })
        return edges
    
    def _generate_transcript_protein_edges(self):
        protein = self._gtf.filter(self._gtf["protein_id"] != "")
        protein_subset = protein.select(['transcript_id', 'protein_id']).unique()
        edges = []
        for t in protein_subset.iter_rows(named=True):
            edges.append({
                "from": {
                    "uuid": t["transcript_id"],
                },
                "to": {
                    "uuid": t["protein_id"],
                },
                "label": "has translation"
            })
        return edges

    def iterate_edges(self):
        if self._gtfloaded == False:
            print("Run iterate_nodes() to load in the gtf file, prior to this function being callable")
            return
        gene_genomic_edges = self._generate_genome_to_genomic_interval_edges()
        genomic_coordinate_edges = self._generate_genomic_interval_edges()
        gene_transcript_edges = self._generate_gene_transcript_edges()
        transcript_protein_edges = self._generate_transcript_protein_edges()
        allEdges = [gene_genomic_edges, genomic_coordinate_edges, gene_transcript_edges, transcript_protein_edges]
        self.edges = list(itertools.chain(*allEdges))
        return self.edges

    def process_item(self, item):
        """Processes a single item (node or edge)."""
        # Customize how you want to process each item, e.g., add extra information, filter, etc.
        return item  # In this example, we just return the item as-is.

    def describe_node_properties(self):
        pass

    def describe_edge_properties(self):
        pass
    
    def list_schema(self):
        return None

    def write_serialized_data_to_file(self, directory=""):
        node_filepath = os.path.join(directory, self.node_filename)
        edge_filepath = os.path.join(directory, self.edge_filename)

        with gzip.open(node_filepath, "at") as f:
            for x in self.nodes:
                json.dump(x, f)
                f.write("\n")

        with gzip.open(edge_filepath, "at") as f:
            for x in self.edges:
                json.dump(x, f)
                f.write("\n")



class Edge:
    def __init__(self, label, src=None, trg=None):
        self.label = label
        self.src = src
        self.trg = trg

    def serialize(self, format="data-pack"):
        pass

    def _can_serialize(self):
        if self.src is None:
            raise Exception("Edge Source is missing")


    def _serialize_as_data_pack(self):
        pass


from __future__ import annotations 
from datetime import (
    datetime,
    date
)
from decimal import Decimal 
from enum import Enum 
import re
import sys
from pydantic.version import VERSION  as PYDANTIC_VERSION 
from typing import (
    Any,
    List,
    Literal,
    Dict,
    Optional,
    Union,
    Generic,
    Iterable,
    TypeVar,
    get_args
)
if sys.version_info.minor > 8:
    from typing import Annotated 
else:
    from typing_extensions import Annotated 

if int(PYDANTIC_VERSION[0])>=2:
    from pydantic import (
        BaseModel,
        ConfigDict,
        Field,
        field_validator
    )
else:
    from pydantic import (
        BaseModel,
        Field,
        validator
    )

from pydantic import GetCoreSchemaHandler 
from pydantic_core import (
    CoreSchema,
    core_schema
)
metamodel_version = "None"
version = "None"


class ConfiguredBaseModel(BaseModel):
    model_config = ConfigDict(
        validate_assignment = True,
        validate_default = True,
        extra = "forbid",
        arbitrary_types_allowed = True,
        use_enum_values = True,
        strict = False,
    )
    pass


_T = TypeVar("_T")

_RecursiveListType = Iterable[Union[_T, Iterable["_RecursiveListType"]]]

class AnyShapeArrayType(Generic[_T]):
    @classmethod
    def __get_pydantic_core_schema__(cls, source_type: Any, handler: GetCoreSchemaHandler) -> CoreSchema:
        # double-nested parameterized types here
        # source_type: List[Union[T,List[...]]]
        item_type = Any if get_args(get_args(source_type)[0])[0] is _T else get_args(get_args(source_type)[0])[0]

        item_schema = handler.generate_schema(item_type)
        if item_schema.get("type", "any") != "any":
            item_schema["strict"] = True

        if item_type is Any:
            # Before python 3.11, `Any` type was a special object without a __name__
            item_name = "Any"
        else:
            item_name = item_type.__name__

        array_ref = f"any-shape-array-{item_name}"

        schema = core_schema.definitions_schema(
            core_schema.list_schema(core_schema.definition_reference_schema(array_ref)),
            [
                core_schema.union_schema(
                    [
                        core_schema.list_schema(core_schema.definition_reference_schema(array_ref)),
                        item_schema,
                    ],
                    ref=array_ref,
                )
            ],
        )

        return schema


AnyShapeArray = Annotated[_RecursiveListType, AnyShapeArrayType]

class Object(ConfiguredBaseModel):
    uuid: Optional[str] = Field(None)
    displayName: str = Field(...)
    description: Optional[str] = Field(None)
    dateCreated: Optional[datetime ] = Field(None)
    dateUpdated: Optional[datetime ] = Field(None)


class Node(ConfiguredBaseModel):
    id: str = Field(...)
    labels: Optional[AnyShapeArray[str]] = Field(None)


class Edge(ConfiguredBaseModel):
    from_: Object = Field(...)
    to: Object = Field(...)
    label: str = Field(...)


class Sample(Object):
    uuid: Optional[str] = Field(None)
    displayName: str = Field(...)
    description: Optional[str] = Field(None)
    dateCreated: Optional[datetime ] = Field(None)
    dateUpdated: Optional[datetime ] = Field(None)


class Donor(Object):
    uuid: Optional[str] = Field(None)
    displayName: str = Field(...)
    description: Optional[str] = Field(None)
    dateCreated: Optional[datetime ] = Field(None)
    dateUpdated: Optional[datetime ] = Field(None)


class Experiment(Object):
    uuid: Optional[str] = Field(None)
    displayName: str = Field(...)
    description: Optional[str] = Field(None)
    dateCreated: Optional[datetime ] = Field(None)
    dateUpdated: Optional[datetime ] = Field(None)


class EpigeneticExperiment(Experiment):
    uuid: Optional[str] = Field(None)
    displayName: str = Field(...)
    description: Optional[str] = Field(None)
    dateCreated: Optional[datetime ] = Field(None)
    dateUpdated: Optional[datetime ] = Field(None)


class Genome(Object):
    assembly: Optional[str] = Field(None)
    taxon: int = Field(...)
    species: str = Field(..., description="""Is the full species name is lower case (e.g. homo sapiens)""")
    uuid: Optional[str] = Field(None)
    displayName: str = Field(...)
    description: Optional[str] = Field(None)
    dateCreated: Optional[datetime ] = Field(None)
    dateUpdated: Optional[datetime ] = Field(None)


class GenomicInterval(Object):
    chr: Optional[str] = Field(None, description="""Name of the chromosome (or contig, scaffold, etc.)""")
    start: Optional[int] = Field(None, description="""The starting position of the feature in the chromosome or scaffold. The first base in a chromosome is numbered 1""")
    end: Optional[int] = Field(None, description="""The ending position of the feature in the chromosome or scaffold. The chromEnd base is not included in the display of the feature. For example, the first 100 bases of a chromosome are defined as chromStart=1, chromEnd=100,  and span the bases numbered 1-100.""")
    taxon: int = Field(...)
    species: str = Field(..., description="""Is the full species name is lower case (e.g. homo sapiens)""")
    uuid: Optional[str] = Field(None)
    displayName: str = Field(...)
    description: Optional[str] = Field(None)
    dateCreated: Optional[datetime ] = Field(None)
    dateUpdated: Optional[datetime ] = Field(None)


class Gene(Object):
    """
    A region (or regions) that includes all of the sequence elements necessary to encode a functional transcript.  A gene may include regulatory regions, transcribed regions and/or other functional sequence regions.
    """
    chr: Optional[str] = Field(None, description="""Name of the chromosome (or contig, scaffold, etc.)""")
    start: Optional[int] = Field(None, description="""The starting position of the feature in the chromosome or scaffold. The first base in a chromosome is numbered 1""")
    end: Optional[int] = Field(None, description="""The ending position of the feature in the chromosome or scaffold. The chromEnd base is not included in the display of the feature. For example, the first 100 bases of a chromosome are defined as chromStart=1, chromEnd=100,  and span the bases numbered 1-100.""")
    assembly: Optional[str] = Field(None)
    taxon: int = Field(...)
    strand: Optional[str] = Field(None, description="""+/- to denote strand or orientation (whenever applicable). Use \".\" if no orientation is assigned.""")
    uuid: Optional[str] = Field(None)
    displayName: str = Field(...)
    description: Optional[str] = Field(None)
    dateCreated: Optional[datetime ] = Field(None)
    dateUpdated: Optional[datetime ] = Field(None)


class Transcript(Object):
    chr: Optional[str] = Field(None, description="""Name of the chromosome (or contig, scaffold, etc.)""")
    start: Optional[int] = Field(None, description="""The starting position of the feature in the chromosome or scaffold. The first base in a chromosome is numbered 1""")
    end: Optional[int] = Field(None, description="""The ending position of the feature in the chromosome or scaffold. The chromEnd base is not included in the display of the feature. For example, the first 100 bases of a chromosome are defined as chromStart=1, chromEnd=100,  and span the bases numbered 1-100.""")
    assembly: Optional[str] = Field(None)
    taxon: int = Field(...)
    uuid: Optional[str] = Field(None)
    displayName: str = Field(...)
    description: Optional[str] = Field(None)
    dateCreated: Optional[datetime ] = Field(None)
    dateUpdated: Optional[datetime ] = Field(None)


class Protein(Object):
    assembly: Optional[str] = Field(None)
    taxon: int = Field(...)
    uuid: Optional[str] = Field(None)
    displayName: str = Field(...)
    description: Optional[str] = Field(None)
    dateCreated: Optional[datetime ] = Field(None)
    dateUpdated: Optional[datetime ] = Field(None)


class GenomeContainsInterval(Edge):
    from_: Genome = Field(...)
    to: GenomicInterval = Field(...)
    label: str = Field(...)


class HasTranslation(Edge):
    from_: Transcript = Field(...)
    to: Protein = Field(...)
    label: str = Field(...)


class TranscribedTo(Edge):
    from_: Gene = Field(...)
    to: Transcript = Field(...)
    label: str = Field(...)


# Model rebuild
# see https://pydantic-docs.helpmanual.io/usage/models/#rebuilding-a-model
Object.model_rebuild()
Node.model_rebuild()
Edge.model_rebuild()
Sample.model_rebuild()
Donor.model_rebuild()
Experiment.model_rebuild()
EpigeneticExperiment.model_rebuild()
Genome.model_rebuild()
GenomicInterval.model_rebuild()
Gene.model_rebuild()
Transcript.model_rebuild()
Protein.model_rebuild()
GenomeContainsInterval.model_rebuild()
HasTranslation.model_rebuild()
TranscribedTo.model_rebuild()




from biobox_analytics.data.adapters._base import Adapter
import _structs as structs
import rdflib
import json
import requests


class GoAdapter(Adapter):
    def pull_data(self):
        pass

    def iterate_nodes(self):
        pass

    def iterate_edges(self):
        pass

    def process_item(self, item):
        """Processes a single item (node or edge)."""
        # Customize how you want to process each item, e.g., add extra information, filter, etc.
        return item  # In this example, we just return the item as-is.

    def describe_node_properties(self):
        pass

    def describe_edge_properties(self):
        pass



from __future__ import annotations
from datetime import (
    datetime,
    date
)
from decimal import Decimal
from enum import Enum
import re
import sys
from pydantic.version import VERSION as PYDANTIC_VERSION

if int(PYDANTIC_VERSION[0]) >= 2:
    from pydantic import (
        BaseModel,
        ConfigDict,
        Field,
        field_validator
    )
else:
    from pydantic import (
        BaseModel,
        Field,
        validator
    )

from typing import (
    Any,
    List,
    Literal,
    Dict,
    Optional,
    Union,
    Generic,
    TypeVar,
    _GenericAlias
)

metamodel_version = "None"
version = "None"


class WeakRefShimBaseModel(BaseModel):
    __slots__ = '__weakref__'


class ConfiguredBaseModel(WeakRefShimBaseModel,
                          validate_assignment=True,
                          validate_all=True,
                          underscore_attrs_are_private=True,
                          extra="forbid",
                          arbitrary_types_allowed=True,
                          use_enum_values=True):
    pass


class ModificationTypes(str, Enum):
    PTM = "PTM"


_T = TypeVar("_T")


class AnyShapeArray(Generic[_T]):
    type_: Type[Any] = Any

    def __class_getitem__(cls, item):
        alias = type(f"AnyShape_{str(item.__name__)}", (AnyShapeArray,), {"type_": item})
        alias.type_ = item
        return alias

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def __modify_schema__(cls, field_schema):
        try:
            item_type = field_schema["allOf"][0]["type"]
            type_schema = {"type": item_type}
            del field_schema["allOf"]
        except KeyError as e:
            if "allOf" in str(e):
                item_type = "Any"
                type_schema = {}
            else:
                raise e

        array_id = f"#any-shape-array-{item_type}"
        field_schema["anyOf"] = [
            type_schema,
            {"type": "array", "items": {"$ref": array_id}},
        ]
        field_schema["$id"] = array_id

    @classmethod
    def validate(cls, v: Union[List[_T], list]):
        if str(type(v)) == "<class 'numpy.ndarray'>":
            v = v.tolist()

        if not isinstance(v, list):
            raise TypeError(f"Must be a list of lists! got {v}")

        def _validate(_v: Union[List[_T], list]):
            for item in _v:
                if isinstance(item, list):
                    _validate(item)
                else:
                    try:
                        anytype = cls.type_.__name__ in ("AnyType", "Any")
                    except AttributeError:
                        # in python 3.8 and 3.9, `typing.Any` has no __name__
                        anytype = str(cls.type_).split(".")[-1] in ("AnyType", "Any")

                    if not anytype and not isinstance(item, cls.type_):
                        raise TypeError(
                            (
                                f"List items must be list of lists, or the type used in "
                                f"the subscript ({cls.type_}. Got item {item} and outer value {v}"
                            )
                        )
            return _v

        return _validate(v)


class Object(ConfiguredBaseModel):
    uuid: Optional[str] = Field(None)
    displayName: str = Field(...)
    description: Optional[str] = Field(None)
    dateCreated: Optional[datetime] = Field(None)
    dateUpdated: Optional[datetime] = Field(None)


class Node(ConfiguredBaseModel):
    _id: str = Field(...)
    labels: Optional[AnyShapeArray[str]] = Field(None)


class Edge(ConfiguredBaseModel):
    _from: Object = Field(...)
    to: Object = Field(...)
    label: str = Field(...)


class Sample(Object):
    uuid: Optional[str] = Field(None)
    displayName: str = Field(...)
    description: Optional[str] = Field(None)
    dateCreated: Optional[datetime] = Field(None)
    dateUpdated: Optional[datetime] = Field(None)


class Donor(Object):
    uuid: Optional[str] = Field(None)
    displayName: str = Field(...)
    description: Optional[str] = Field(None)
    dateCreated: Optional[datetime] = Field(None)
    dateUpdated: Optional[datetime] = Field(None)


class Experiment(Object):
    uuid: Optional[str] = Field(None)
    displayName: str = Field(...)
    description: Optional[str] = Field(None)
    dateCreated: Optional[datetime] = Field(None)
    dateUpdated: Optional[datetime] = Field(None)


class EpigeneticExperiment(Experiment):
    uuid: Optional[str] = Field(None)
    displayName: str = Field(...)
    description: Optional[str] = Field(None)
    dateCreated: Optional[datetime] = Field(None)
    dateUpdated: Optional[datetime] = Field(None)


class Genome(Object):
    assembly: Optional[str] = Field(None)
    taxon: int = Field(...)
    uuid: Optional[str] = Field(None)
    displayName: str = Field(...)
    description: Optional[str] = Field(None)
    dateCreated: Optional[datetime] = Field(None)
    dateUpdated: Optional[datetime] = Field(None)


class GenomicInterval(Object):
    chr: Optional[str] = Field(None, description="""Name of the chromosome (or contig, scaffold, etc.)""")
    start: Optional[int] = Field(None,
                                 description="""The starting position of the feature in the chromosome or scaffold. The first base in a chromosome is numbered 0""")
    end: Optional[int] = Field(None,
                               description="""The ending position of the feature in the chromosome or scaffold. The chromEnd base is not included in the display of the feature. For example, the first 100 bases of a chromosome are defined as chromStart=0, chromEnd=100,  and span the bases numbered 0-99.""")
    taxon: int = Field(...)
    species: str = Field(..., description="""Is the full species name is lower case (e.g. homo sapiens)""")
    uuid: Optional[str] = Field(None)
    displayName: str = Field(...)
    description: Optional[str] = Field(None)
    dateCreated: Optional[datetime] = Field(None)
    dateUpdated: Optional[datetime] = Field(None)


class Gene(Object):
    """
    A region (or regions) that includes all of the sequence elements necessary to encode a functional transcript.  A gene may include regulatory regions, transcribed regions and/or other functional sequence regions.
    """
    uuid: Optional[str] = Field(None)
    displayName: str = Field(...)
    description: Optional[str] = Field(None)
    dateCreated: Optional[datetime] = Field(None)
    dateUpdated: Optional[datetime] = Field(None)


class Transcript(Object):
    uuid: Optional[str] = Field(None)
    displayName: str = Field(...)
    description: Optional[str] = Field(None)
    dateCreated: Optional[datetime] = Field(None)
    dateUpdated: Optional[datetime] = Field(None)


class Protein(Object):
    uuid: Optional[str] = Field(None)
    displayName: str = Field(...)
    description: Optional[str] = Field(None)
    dateCreated: Optional[datetime] = Field(None)
    dateUpdated: Optional[datetime] = Field(None)


class HasTranslation(Edge):
    _from: Transcript = Field(...)
    to: Protein = Field(...)
    label: str = Field(...)


class ChipSeq(EpigeneticExperiment):
    """
    A method to determine the genomic regions that proteins bind to.
    """
    uuid: Optional[str] = Field(None)
    displayName: str = Field(...)
    description: Optional[str] = Field(None)
    dateCreated: Optional[datetime] = Field(None)
    dateUpdated: Optional[datetime] = Field(None)


class ModifiedProtein(Protein):
    """
    A protein that has a post-translational modification.
    """
    modification_type: Optional[ModificationTypes] = Field(None)
    uuid: Optional[str] = Field(None)
    displayName: str = Field(...)
    description: Optional[str] = Field(None)
    dateCreated: Optional[datetime] = Field(None)
    dateUpdated: Optional[datetime] = Field(None)


class NarrowPeakOn(Edge):
    """
    Peaks of signal enrichment based on pooled, normalized (interpreted) data. It is a BED6+4 format.
    """
    chr: Optional[str] = Field(None, description="""Name of the chromosome (or contig, scaffold, etc.)""")
    start: Optional[int] = Field(None,
                                 description="""The starting position of the feature in the chromosome or scaffold. The first base in a chromosome is numbered 0""")
    end: Optional[int] = Field(None,
                               description="""The ending position of the feature in the chromosome or scaffold. The chromEnd base is not included in the display of the feature. For example, the first 100 bases of a chromosome are defined as chromStart=0, chromEnd=100,  and span the bases numbered 0-99.""")
    strand: Optional[str] = Field(None,
                                  description="""+/- to denote strand or orientation (whenever applicable). Use \".\" if no orientation is assigned.""")
    score: Optional[int] = Field(None,
                                 description="""Indicates how dark the peak will be displayed in the browser (0-1000).  If all scores were \"'0\"' when the data were submitted to the DCC,  the DCC assigned scores 1-1000 based on signal value.  Ideally the average signalValue per base spread is between 100-1000.""")
    signalValue: Optional[float] = Field(None,
                                         description="""Measurement of overall (usually, average) enrichment for the region.""")
    pValue: Optional[float] = Field(None,
                                    description="""Measurement of statistical significance (-log10). Use -1 if no pValue is assigned.""")
    qValue: Optional[float] = Field(None,
                                    description="""Measurement of statistical significance using false discovery rate (-log10). Use -1 if no qValue is assigned.""")
    peak: Optional[int] = Field(None,
                                description="""Point-source called for this peak; 0-based offset _from chromStart. Use -1 if no point-source called.""")
    _from: EpigeneticExperiment = Field(...)
    to: GenomicInterval = Field(...)
    label: str = Field(...)


class IsAModifiedFormOf(Edge):
    """
    A gene has a translation to a protein.
    """
    _from: ModifiedProtein = Field(...)
    to: Protein = Field(...)
    label: str = Field(...)


class AssayTargetOn(Edge):
    """
    Describes the target for the immuno-precipitation and/or pull down assay.
    """
    _from: ChipSeq = Field(...)
    to: Protein = Field(...)
    label: str = Field(...)


# Update forward refs
# see https://pydantic-docs.helpmanual.io/usage/postponed_annotations/
Object.update_forward_refs()
Node.update_forward_refs()
Edge.update_forward_refs()
Sample.update_forward_refs()
Donor.update_forward_refs()
Experiment.update_forward_refs()
EpigeneticExperiment.update_forward_refs()
Genome.update_forward_refs()
GenomicInterval.update_forward_refs()
Gene.update_forward_refs()
Transcript.update_forward_refs()
Protein.update_forward_refs()
HasTranslation.update_forward_refs()
ChipSeq.update_forward_refs()
ModifiedProtein.update_forward_refs()
NarrowPeakOn.update_forward_refs()
IsAModifiedFormOf.update_forward_refs()
AssayTargetOn.update_forward_refs()
